import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from dataiku import insights


def bin_dataframe(df, output_col, target_column, variables, num_bins=10):
    df = df[variables + [output_col,target_column]].copy()
    global_nok_rate = 100 * df['defect_binary'].value_counts()['NOK']/(df['defect_binary'].value_counts()['OK']+df['defect_binary'].value_counts()['NOK'])

    for col in df.columns:

        if df[col].dtype.kind in 'biufc':  # The column is a numerical variable
            # If the column has only 2 unique values, handle it as binary
            if df[col].nunique() == 2:
                continue

            # Calculate the IQR
            Q1 = df[col].quantile(0.25)
            Q3 = df[col].quantile(0.75)
            IQR = Q3 - Q1

            # Filter outliers
            filtered_df = df.loc[(df[col] >= Q1 - 1.5 * IQR) & (df[col] <= Q3 + 1.5 * IQR)].copy()

            # Determine the bins from the filtered data
            _, bins = np.histogram(filtered_df[col], bins=num_bins)

            # Cut the original dataframe column using the bins
            df[col] = pd.cut(df[col], bins=bins, include_lowest=True)

    return df, global_nok_rate

def calculate_defect_rate(df, output_col, global_nok_rate):
    impact_df = pd.DataFrame(columns=['variable', 'bin', 'nok_rate', 'points_in_bin', 'impact', 'color'])

    for col in df.columns:
        if col != output_col:
            if df[col].dtype.kind in 'O':  # Categorical variable
                impacts = []
                for cat in df[col].dropna().unique():  # Drop NaN values to prevent NaN bins
                    slice_df = df[df[col] == cat]
                    bin_nok_rate = (slice_df[output_col] == 'NOK').mean() * 100
                    score = np.abs(bin_nok_rate - global_nok_rate) * (len(slice_df) / len(df[col].dropna()))
                    color = 'green' if bin_nok_rate <= global_nok_rate else 'red'

                    impacts.append({
                        'variable': col,
                        'bin': cat,
                        'nok_rate': bin_nok_rate,
                        'points_in_bin': len(slice_df),
                        'impact': score,
                        'color': color
                    })
                impact_df = pd.concat([impact_df, pd.DataFrame(impacts)], ignore_index=True)
            else:
                df_num = df[df[col].notna()]  # Exclude rows where the column has NaN values
                for bin_range in df_num[col].unique():
                    slice_df = df_num[df_num[col] == bin_range]
                    
                    # Avoid processing empty bins
                    if slice_df.empty:
                        continue
                    
                    bin_nok_rate = (slice_df[output_col] == 'NOK').mean() * 100
                    score = np.abs(bin_nok_rate - global_nok_rate) * (len(slice_df) / len(df[col].dropna()))
                    color = 'green' if bin_nok_rate <= global_nok_rate else 'red'
                                                                              
                    impacts = {
                        'variable': col,
                        'bin': bin_range,
                        'nok_rate': bin_nok_rate,
                        'points_in_bin': len(slice_df),
                        'impact': score,
                        'color': color
                    }
                    impact_df = impact_df.append(impacts, ignore_index=True)

    variable_impact = impact_df.groupby('variable')['impact'].apply(lambda x: np.abs(x).sum())
    
    # Convert to DataFrame and rename columns
    variable_impact_df = variable_impact.reset_index()
    variable_impact_df.columns = ['variable', 'relative_score']

    # Normalize the relative_score
    max_score = variable_impact_df['relative_score'].max()
    variable_impact_df['relative_score'] = (variable_impact_df['relative_score'] / max_score) * 100

    variable_impact_df['rank'] = variable_impact_df['relative_score'].rank(ascending=False, method='min')
    variable_impact_df = variable_impact_df.sort_values('rank')
    
    return impact_df, variable_impact_df



def calculate_zone_impact(variable_df, global_nok_rate):
    """
    Calculate zone impact based on the condition if the defect_rate is below the average defect rate.
    Zones are groups of adjacent bins with the same color.
    """
    most_impacting_green = None
    prev_color = 'green' if variable_df['nok_rate'].values[0] <= global_nok_rate else 'red'
    start_idx = 0
    max_impact_green = 0

    for idx in range(1, len(variable_df)):
        current_color = 'green' if variable_df['nok_rate'].values[idx] <= global_nok_rate else 'red'
        if current_color != prev_color:
            impact_sum = variable_df['impact'].values[start_idx:idx].sum()
            if prev_color == 'green':
                if impact_sum > max_impact_green:
                    max_impact_green = impact_sum
                    most_impacting_green = (start_idx, idx-1, prev_color)

            start_idx = idx
            prev_color = current_color

    # Check the last zone
    impact_sum = variable_df['impact'].values[start_idx:].sum()
    if prev_color == 'green':
        if impact_sum > max_impact_green:
            most_impacting_green = (start_idx, len(variable_df) - 1, prev_color)

    most_impacting_zones = []
    if most_impacting_green and most_impacting_green[1] - most_impacting_green[0] >= 1:
        most_impacting_zones.append(most_impacting_green)

    return most_impacting_zones


def generate_impact_figure(impact_df, variable_impact, global_nok_rate):
    for i in range(0, 30):
        fig, ax1 = plt.subplots(figsize=(16, 10), dpi=90)
        plt.subplots_adjust(left=0.1, right=0.90, top=0.9, bottom=0.1)

        # Initialize zone_impacts to an empty list for each iteration
        zone_impacts = []

        if i < len(variable_impact.index):
            variable = variable_impact['variable'].iloc[i]
            variable_df = impact_df[impact_df['variable'] == variable]

            # Processing for numerical data
            if isinstance(variable_df['bin'].iloc[0], pd.Interval):
                variable_df = variable_df.sort_values(by='bin')
                bin_midpoints = [interval.mid for interval in variable_df['bin']]
                sorted_midpoints = np.sort(bin_midpoints)
                smallest_diff = np.min(np.diff(sorted_midpoints))
                grey_bar_width = 0.8 * smallest_diff
                green_red_bar_width = 0.8 * grey_bar_width
                left_bound = variable_df['bin'].apply(lambda x: x.left).min()
                right_bound = variable_df['bin'].apply(lambda x: x.right).max()

                # Calculate zones for numerical data
                zone_impacts = calculate_zone_impact(variable_df, global_nok_rate)
                
                # Add shading for the zones
                for start_idx, end_idx, color in zone_impacts:
                    if color == 'green':
                        if isinstance(variable_df['bin'].iloc[0], pd.Interval):
                            left_value = variable_df['bin'].values[start_idx].left
                            right_value = variable_df['bin'].values[end_idx].right
                            ax1.axvspan(left_value, right_value, alpha=0.5, color='darkseagreen', zorder=1)
                        elif isinstance(variable_df['bin'].iloc[0], str):
                            ax1.axvspan(start_idx - 0.4, end_idx + 0.4, alpha=0.5, color='darkseagreen', zorder=1)
                
            # Processing for categorical data
            elif isinstance(variable_df['bin'].iloc[0], (str, int, float)):
                variable_df['bin'] = pd.Categorical(variable_df['bin'], ordered=True, categories=sorted(variable_df['bin'].unique()))
                variable_df = variable_df.sort_values(by='bin')
                bin_midpoints = variable_df['bin'].cat.codes  # Use category codes as midpoints
                sorted_midpoints = bin_midpoints.copy()
                ax1.set_xticks(sorted_midpoints)  # set xticks position
                ax1.set_xticklabels(variable_df['bin'].cat.categories)  # set xticks labels

                smallest_diff = 1  # Each category is considered as one unit apart
                grey_bar_width = 0.8 * smallest_diff
                green_red_bar_width = 0.8 * grey_bar_width
                left_bound = 0
                right_bound = len(variable_df['bin'].cat.categories) - 1
                
                # Calculate zones for categorical data
                zone_impacts = calculate_zone_impact(variable_df, global_nok_rate)
                
                # Add shading for the zones
                for start_idx, end_idx, color in zone_impacts:
                    if color == 'green':
                        if isinstance(variable_df['bin'].iloc[0], pd.Interval):
                            left_value = variable_df['bin'].values[start_idx].left
                            right_value = variable_df['bin'].values[end_idx].right
                            ax1.axvspan(left_value, right_value, alpha=0.5, color='darkseagreen', zorder=1)
                        elif isinstance(variable_df['bin'].iloc[0], str):
                            ax1.axvspan(start_idx - 0.4, end_idx + 0.4, alpha=0.5, color='darkseagreen', zorder=1)
                            
            
            # Plotting bars and labels
            ax1.bar(bin_midpoints, variable_df['points_in_bin'], color='gray', alpha=0.5, width=grey_bar_width, zorder=2)
            ax1.set_xlabel(variable, fontsize=18)
            ax1.set_ylabel("Number of Points", fontsize=18)
            ax1.tick_params(axis='both', labelsize=10)

            ax2 = ax1.twinx()
            ax2.yaxis.set_label_position("right")
            ax2.yaxis.tick_right()
            ax2.set_ylabel("Defect Rate (%)", fontsize=18)
            ax2.tick_params(axis='both', labelsize=16)

            # Fetch colors directly from the 'impact' column in the variable_df
            colors = variable_df['color'].tolist()

            
            bars = ax2.bar(bin_midpoints, variable_df['nok_rate'] - global_nok_rate,
                           bottom=global_nok_rate,
                           color=colors,
                           width=green_red_bar_width, zorder=2)

            for bar, rate in zip(bars, variable_df['nok_rate']):
                height = bar.get_height()
                ax2.text(bar.get_x() + bar.get_width() / 2, bar.get_y() + height / 2,
                         f'{rate:.2f}%', ha='center', va='center', color='black', fontname='DejaVu Sans Mono', fontsize=14)

            ax2.plot([left_bound, right_bound], [global_nok_rate] * 2, 'k--')

            plt.title(f"{variable}", fontsize=20, pad=30)
            plt.text(0.5, 1.02, f"Defect Rate: {global_nok_rate:.2f}%    Rank: {variable_impact[variable_impact['variable'] == variable]['rank'].values[0]:.0f}    Impact: {variable_impact[variable_impact['variable'] == variable]['relative_score'].values[0]:.2f}%",
                     horizontalalignment='center', verticalalignment='center',
                     transform=ax1.transAxes, fontsize=18)

            # Process the zone impacts
            for start_idx, end_idx, color in zone_impacts:
                if color == 'green':
                    top_y_value = ax1.get_ylim()[1] * 0.95  # Near the top of the y-axis for vertical placement

                    # Handle numerical data
                    if isinstance(variable_df['bin'].iloc[0], pd.Interval):
                        left_value = variable_df['bin'].values[start_idx].left
                        right_value = variable_df['bin'].values[end_idx].right
                        ax1.text(left_value, top_y_value, f"{left_value:.2f}", ha='left', va='center', color='#008000', fontsize=18, weight='bold')
                        ax1.text(right_value, top_y_value, f"{right_value:.2f}", ha='right', va='center', color='#008000', fontsize=18, weight='bold')

                    # Handle categorical data
                    elif isinstance(variable_df['bin'].iloc[0], str):
                        if start_idx < len(variable_df) and end_idx < len(variable_df):
                            left_value = variable_df['bin'].values[start_idx]
                            right_value = variable_df['bin'].values[end_idx]
                            ax1.text(start_idx, top_y_value, f"{left_value}", ha='left', va='center', color='#008000', fontsize=18, weight='bold')
                            ax1.text(end_idx, top_y_value, f"{right_value}", ha='right', va='center', color='#008000', fontsize=18, weight='bold')
                        else:
                            print(f"Warning: Out-of-bounds indices detected. start_idx: {start_idx}, end_idx: {end_idx}")

        else:
            ax1.axis('off')  # Hide axis
            plt.text(0.5, 0.5, 'No more variables to display', horizontalalignment='center',
                     verticalalignment='center', transform=ax1.transAxes, fontsize=20)

        filename = f'impact_chart_{i:03}'
        # Assuming you have an insights module with save_figure function, otherwise adjust as needed
        insights.save_figure(filename, fig)
        plt.close(fig)


def generate_input_statistics_charts(df, target_column, new_column_name):
    # Create a copy of the DataFrame to keep the original unchanged
    df_copy = df.copy()

    # Donut chart
    plt.figure(figsize=(10, 10), dpi=90)
    ax = plt.gca()
    counts = df_copy[new_column_name].value_counts()
    colors = ['red' if label == 'NOK' else 'green' for label in counts.index]
    def func(pct, allvals):
        absolute = int(pct/100.*np.sum(allvals))
        return "{:.1f}%".format(pct, absolute)
    wedges, texts, autotexts = ax.pie(counts, labels=counts.index, colors=colors, 
                                       autopct=lambda pct: func(pct, counts), 
                                       startangle=140, 
                                       textprops={'fontsize': 14},  # Increase font size
                                       labeldistance=1.1)  # Adjust label distance
    centre_circle = plt.Circle((0, 0), 0.70, fc='white')
    ax.add_artist(centre_circle)
    plt.title('Distribution of ' + target_column, fontsize=20, y=1.05)
    filename = 'input_donut_chart'
    insights.save_figure(filename, plt.gcf())
    plt.close()

    # Map the labels to 1 and 0
    df_copy[new_column_name] = df_copy[new_column_name].map({'NOK': 1, 'OK': 0})
    counts = df_copy[new_column_name].value_counts()

    # Bar chart
    fig, ax = plt.subplots(figsize=(12, 8), dpi=90)
    for value in counts.index:
        ax.bar(value, counts.loc[value], color='red' if value == 1 else 'green')
        ax.text(value, counts.loc[value], counts.loc[value], ha='center', va='bottom')
    ax.set_title('Count', fontsize=16)

    # Reduce margins
    plt.subplots_adjust(left=0.1, right=0.9, top=0.9, bottom=0.1)

    filename = 'input_bar_chart'
    insights.save_figure(filename, fig)
    plt.close(fig)

    # Statistics table
    rows = ['Number of Points:', 'Min:', 'Max:', 'Average:', 'Std. Dev:']

    cell_text = [
        [str(int(df_copy[new_column_name].count()))],
        [f"{df_copy[new_column_name].min():.2f}"],
        [f"{df_copy[new_column_name].max():.2f}"],
        [f"{df_copy[new_column_name].mean():.2f}"],
        [f"{df_copy[new_column_name].std():.2f}"]
    ]

    df_stats = pd.DataFrame(cell_text, index=rows, columns=[''])

    # You can apply additional styling as needed:
    styled_df = df_stats.style.set_table_styles([
        dict(selector="th, td", props=[("text-align", "left"), ("font-family", "Verdana")])
    ])

    html = styled_df.render()

    filename = 'input_stats_table'
    insights.save_data(filename, html, content_type="text/html")