import re
from collections import defaultdict
from datetime import datetime


def is_plausible_date(value):
    """
    Checks if the given value is a plausible date string.
    """
    if isinstance(value, str) and 6 <= len(value) <= 25:
        if any(char in value for char in ['-', '/', ' ']):
            return True
    return False


def find_date_format(date_str):
    format_patterns = {
        r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}': 'yyyy-MM-dd HH:mm:SS',
        r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}': "yyyy-MM-dd'T'HH:mm:ss",
        r'\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}': 'dd-MM-yyyy HH:mm:SS',
        r'\d{4}/\d{2}/\d{2}': 'yyyy/MM/dd',
        r'\d{2}/\d{2}/\d{4}': 'dd/MM/yyyy',
        r'\d{2}-\d{2}-\d{4}': 'dd-MM-yyyy',
        r'\d{2}-\d{2}-\d{2}': 'dd-MM-yy'
    }

    for pattern, format in format_patterns.items():
        if re.match(pattern, date_str):
            return format
    return None


def detect_date_columns(df):
    date_columns = []
    for column in df.columns:
        if df[column].dtype == 'object':  # String columns
            non_empty_values = df[column].dropna()
            plausible_dates = non_empty_values.apply(is_plausible_date)
            if plausible_dates.sum() > 0.90 * len(non_empty_values):
                format_counter = defaultdict(int)
                for date_str in non_empty_values[plausible_dates]:
                    date_format = find_date_format(date_str)
                    if date_format:
                        format_counter[date_format] += 1
                most_common_format = max(format_counter, key=format_counter.get, default=None)
                if most_common_format:
                    date_columns.append({'name': column, 'format': most_common_format})
    return date_columns


def convert_date_format(date_str):
    format_patterns = {
            r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}': '%Y-%m-%d %H:%M:%S',
            r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}': "%Y-%m-%dT%H:%M:%S",
            r'\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}': '%d-%m-%Y %H:%M:%S',
            r'\d{4}/\d{2}/\d{2}': '%Y/%m/%d',
            r'\d{2}/\d{2}/\d{4}': '%d/%m/%Y',
            r'\d{2}-\d{2}-\d{4}': '%d-%m-%Y',
            r'\d{2}-\d{2}-\d{2}': '%d-%m-%y'
        }

    detected_format = None
    for pattern, python_format_code in format_patterns.items():
        if re.match(pattern, date_str):
            detected_format = python_format_code
            break

    if detected_format == "%Y-%m-%dT%H:%M:%S":
        return date_str

    if detected_format:
        try:
            # Parse the date string to a datetime object using the detected format
            parsed_date = datetime.strptime(date_str, detected_format)
            # If you need to format it differently, do so here
            iso_formatted_date = parsed_date.strftime('%Y-%m-%dT%H:%M:%S')
            return iso_formatted_date
        except ValueError as e:
            print(f"Failed to parse the date with the detected format due to: {e}")
            return date_str
    else:
        return date_str
