from flask import Blueprint, jsonify, request
import json
import pandas as pd
import numpy as np
from .backend_utilities import AnalysisChartFetcher

fetch_api = Blueprint("fetch_api", __name__, url_prefix="/api")

acf = AnalysisChartFetcher()


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, pd.Timestamp):
            return obj.strftime('%Y-%m-%d %H:%M:%S')

        if isinstance(obj, np.int64):
            return int(obj)
        return super().default(obj)


@fetch_api.route("/target_variables", methods=["GET"])
def target_variables():
    targets = acf.get_target_variables()
    return json.dumps(targets)


@fetch_api.route('/target_range', methods=["POST"])
def get_target_range():
    data = request.get_json(force=True)
    targets = acf.get_variable_range_and_stats(data['dates'], data['target'], data["selectedFilters"], include_stats=True)
    return json.dumps(targets, cls=CustomJSONEncoder)


@fetch_api.route('/filter_range', methods=["POST"])
def get_filter_range():
    data = request.get_json(force=True)
    targets = acf.get_variable_range_and_stats(data['dates'], data['target'], include_stats=False)
    return json.dumps(targets, cls=CustomJSONEncoder)


@fetch_api.route('/create_target_chart', methods=["POST"])
def create_target_chart():
    data = request.get_json(force=True)
    response = acf.get_target_chart_data(data['target'], data['dates'], data['selectedFilters'])
    return json.dumps(response, cls=CustomJSONEncoder)


@fetch_api.route('/update_target_range', methods=["POST"])
def update_target_range():
    data = request.get_json(force=True)
    response = acf.update_target_range(data['target'], data['good_range'], data['selectedFilters'])
    return json.dumps(response, cls=CustomJSONEncoder)


@fetch_api.route("/analysis_variables/<target>", methods=["GET"])
def analysis_variables(target):
    resp = acf.get_analysis_variables(target)
    return json.dumps(resp)


@fetch_api.route('/get_analysis', methods=["POST"])
def get_analysis():
    data = request.get_json(force=True)
    resp = acf.get_analysis_results(data['target'], data['bin'], data["filters"]['variables'], data["filters"]['dates'], data["filters"]["good_range"], data['selectedFilters'])
    return json.dumps(resp)


@fetch_api.route("/target_state", methods=["POST"])
def target_state():
    request_params = request.get_json(force=True)
    target = request_params['target']
    resp = acf.get_target_state(target)
    return json.dumps(resp)


@fetch_api.route('/update_target_state', methods=["POST"])
def update_target_state():
    data = request.get_json(force=True)
    target = data["target"]
    selected_variables_limits = data['variables']
    resp = acf.get_current_state(target, selected_variables_limits)
    return jsonify(resp)


@fetch_api.route("/get_studies", methods=["GET"])
def get_studies():
    resp = acf.get_saved_studies()
    print('got studies:', resp)
    return jsonify(resp)


@fetch_api.route("/check_study/<studyId>", methods=["GET"])
def check_study_copy(studyId):
    resp = acf.check_study_copy(studyId)
    return json.dumps(resp)


@fetch_api.route("/study/<studyId>", methods=["GET"])
def get_study(studyId):
    resp = acf.get_study(studyId)
    print('study:', resp)
    return json.dumps(resp)


@fetch_api.route("/study", methods=["POST"])
def save_study():
    request_params = request.get_json(force=True)
    headers = dict(request.headers)
    resp = acf.save_study(request_params['target_info'], request_params['variables'], request_params['analysis_variables'], request_params['selectedFilters'], headers)
    return jsonify(resp)


@fetch_api.route("/study/<studyId>", methods=["DELETE"])
def delete_study(studyId):
    resp = acf.delete_study(studyId)
    return jsonify(resp)
