import pandas as pd
import numpy as np
from math import isnan, log10, floor
import time
import hashlib
from webaiku.apis.dataiku.improved_dataset import Dataset
from typing import List, Dict, Optional
from datetime import datetime, timedelta
import dataiku
from dataiku import SQLExecutor2
from collections import defaultdict
import traceback
import logging
import concurrent.futures

from dku_utils.projects.recipes.recipe_commons import (
    get_recipe_settings_and_dictionary,
    update_recipe_ouput_schema,
)
from .metrics_handling import add_metrics_probes

numerical_types = {
    "int",
    "tinyint",
    "smallint",
    "bigint",
    "numeric",
    "double precision",
    "double",
    "float",
    "date",
}

supported_connections = ["SQLServer", "Snowflake", "PostgreSQL"]


def manage_defect_binary_col(table_name, target_binary_col_name, dataset_name=None):
    """
    This function checks if a defect_binary col exists in the given SQL table,
    and if not, it adds the column.

    :param table_name: The name of the table to modify.
    :param defect_binary_col_name: The name of the defect binary column to add.
    :param dataset_name: The dataset name for the SQL executor.
    """
    # SQL query to check if the column exists and add it if not
    check_and_add_query = f"""
    DO $$
    BEGIN
        IF NOT EXISTS (
            SELECT 1
            FROM information_schema.columns
            WHERE table_name = '{table_name}' AND column_name = '{target_binary_col_name}'
        ) THEN
            ALTER TABLE "{table_name}" ADD COLUMN "{target_binary_col_name}" VARCHAR(255);
        END IF;
    END $$;
    """

    try:
        # Execute the query
        executor = SQLExecutor2(dataset=dataset_name)
        executor.query_to_df(check_and_add_query, post_queries=["COMMIT"])
    except Exception as e:
        print(f"An error occurred: {str(e)}")


def get_columns_statistics(
    project_name,
    main_connection,
    dataset_name: str,
    date_col_name: str,
    col_list: List[str],
    aggregations: Optional[List[str]] = ["min", "max"],
    date_from: Optional[str] = None,
    date_to: Optional[str] = None,
    selected_filters: Optional[dict] = None,
    verbose: Optional[bool] = True
):

    start_time = time.time()
    # Define numerical and categorical columns
    numerical_cols = [col for col in col_list if col["type"] in numerical_types]
    categorical_cols = [col for col in col_list if col["type"] == "string" or col["type"] == "boolean"]

    # Get the connection type
    client = dataiku.api_client()
    project = client.get_project(project_name)
    connection_type = client.get_connection(main_connection).get_info()["type"]

    # Checking connection type:

    if connection_type in supported_connections:
        statistics = compute_statistics_sql(
            dataset_name,
            main_connection,
            connection_type,
            numerical_cols,
            categorical_cols,
            aggregations,
            date_col_name,
            date_from,
            date_to,
            selected_filters,
            verbose
        )

    else:
        statistics = compute_statistics_dss(
            project,
            dataset_name,
            numerical_cols,
            categorical_cols,
            aggregations,
            date_col_name,
            date_from,
            date_to
        )
    if verbose:
        print(f"[METRICS] get_columns_statistics {round(time.time() - start_time, 3)} seconds")
    return statistics


def compute_statistics_sql(
    dataset_name,
    connection,
    connection_type,
    numerical_cols,
    categorical_cols,
    aggregations,
    date_col_name,
    date_from=None,
    date_to=None,
    selected_filters: Optional[List[Dict]] = None,
    verbose=True
):
    start_time = time.time()

    UNIQUE_SAMPLE_LIMIT = 100000

    # Get the table_name from the dataset_name:
    dataset = dataiku.Dataset(dataset_name)
    location_info = dataset.get_location_info()
    table_name = location_info.get("info", {}).get("table")

    # Construct the SELECT part of the query and corresponding column names
    select_statements = []
    result_column_names = []
    for col in numerical_cols:
        col_name = col["name"]

        for agg in aggregations:
            if agg.lower() == "mean" and connection_type in supported_connections:
                select_statements.append(f'AVG("{col_name}") AS "mean_{col_name}"')
                result_column_names.append(f"mean_{col_name}")

            elif agg.lower() == "median" and connection_type == "PostgreSQL":
                select_statements.append(f'percentile_cont(0.5) WITHIN GROUP (ORDER BY "{col_name}") AS "median_{col_name}"')
                result_column_names.append(f"median_{col_name}")

            elif agg.lower() == "q1" and connection_type in supported_connections:
                select_statements.append(f'percentile_cont(0.25) WITHIN GROUP (ORDER BY "{col_name}") AS "q1_{col_name}"')
                result_column_names.append(f"q1_{col_name}")

            elif agg.lower() == "q3" and connection_type in supported_connections:
                select_statements.append(f'percentile_cont(0.75) WITHIN GROUP (ORDER BY "{col_name}") AS "q3_{col_name}"')
                result_column_names.append(f"q3_{col_name}")

            elif agg.lower() == "stddev":
                if connection_type == "PostgreSQL":
                    select_statements.append(f'STDDEV_SAMP("{col_name}") AS "stddev_{col_name}"')
                elif connection_type == "SQLServer":
                    select_statements.append(f'STDEV("{col_name}") AS "stddev_{col_name}"')
                elif connection_type == "Snowflake":
                    select_statements.append(f'STDDEV("{col_name}") AS "stddev_{col_name}"')
                result_column_names.append(f"stddev_{col_name}")
            else:
                select_statements.append(f'{agg.upper()}("{col_name}") AS "{agg}_{col_name}"')
                result_column_names.append(f"{agg}_{col_name}")

        select_statements.append(f'COUNT("{col_name}") AS "count_{col_name}"')
        result_column_names.append(f"nb_points_{col_name}")

    for col in categorical_cols:
        col_name = col["name"]
        if connection_type == "PostgreSQL":
            mode_subquery = f"""(SELECT "{col_name}" FROM "{table_name}" GROUP BY "{col_name}" ORDER BY COUNT(*) DESC LIMIT 1) AS "mode_{col_name}" """
        elif connection_type == "SQLServer":
            mode_subquery = f"""(SELECT TOP 1 "{col_name}" FROM "{table_name}" GROUP BY "{col_name}" ORDER BY COUNT(*) DESC) AS "mode_{col_name}" """
        elif connection_type == "Snowflake":
            mode_subquery = f"""(SELECT "{col_name}" FROM "{table_name}" GROUP BY "{col_name}" ORDER BY COUNT(*) DESC LIMIT 1) AS "mode_{col_name}" """

        select_statements.append(mode_subquery)
        select_statements.append(f'COUNT(DISTINCT "{col_name}") AS "distinct_{col_name}"')
        select_statements.append(f'COUNT("{col_name}") AS "count_{col_name}"')
        result_column_names.extend([f"mode_{col_name}", f"distinct_{col_name}", f"nb_points_{col_name}"])

    select_clause = ", ".join(select_statements)

    # Constructing the WHERE clause
    where_clause = ""
    where_conditions = []

    if date_from and date_to:
        if connection_type == "PostgreSQL":
            where_conditions.append(f"\"{date_col_name}\" BETWEEN '{date_from}'::date AND '{date_to}'::date")
        elif connection_type == "SQLServer":
            where_conditions.append(f"\"{date_col_name}\" BETWEEN CONVERT(date, '{date_from}') AND CONVERT(date, '{date_to}')")
        elif connection_type == "Snowflake":
            where_conditions.append(f"\"{date_col_name}\" BETWEEN TO_DATE('{date_from}', 'YYYY/MM/DD') AND TO_DATE('{date_to}', 'YYYY/MM/DD')")

    if selected_filters:
        for filter in selected_filters:
            target = filter["target"]
            filter_type = filter["filters"]["type"]
            filter_range = filter["filters"]["range"]

            if filter_type == "cat":
                filter_condition = '"{target}" IN ({values})'.format(
                    target=target,
                    values=", ".join("'{}'".format(value) for value in filter_range)
                )
            elif filter_type == "num":
                filter_condition = f'"{target}" BETWEEN {filter_range[0]} AND {filter_range[1]}'

            where_conditions.append(filter_condition)

    if where_conditions:
        where_clause = "WHERE " + " AND ".join(where_conditions)
    else:
        where_clause = ""

    sql_query = f'SELECT {select_clause} FROM "{table_name}" {where_clause if where_clause else ""}'
    client = dataiku.api_client()
    query = client.sql_query(sql_query, connection=connection)
    results = []
    known_stats = [
        "min",
        "max",
        "mean",
        "stddev",
        "nb_points",
        "q1",
        "median",
        "q3",
        "distinct",
        "mode",
        "unique_values",
    ]

    for row in query.iter_rows():
        stats_dict = {col["name"]: {} for col in numerical_cols + categorical_cols}
        for col_name, value in zip(result_column_names, row):
            stat = next((s for s in known_stats if col_name.startswith(s)), "value")
            var_name = col_name[len(stat):].lstrip("_")
            if var_name in stats_dict:
                # For SQLServer or others, adjust the parsing logic accordingly
                # TODO: FIX for SQL Server
                stats_dict[var_name][stat] = value
        results.append(stats_dict)
    query.verify()

    final_results = {}
    for stat_dict in results:
        for var, stats in stat_dict.items():
            if var not in final_results:
                final_results[var] = {}
            final_results[var].update(stats)

        # Handling the unique_value separately to avoid type casting issues:

        for col in categorical_cols:
            col_name = col["name"]
            unique_select_statement = []
            if connection_type in ["PostgreSQL", "Snowflake"]:
                unique_select_statement.append(
                    f'(SELECT DISTINCT sub."{col_name}"'
                    f'FROM (SELECT "{col_name}" FROM "{table_name}" '
                    f'GROUP BY "{col_name}" LIMIT {UNIQUE_SAMPLE_LIMIT}) AS sub)'
                )

            elif connection_type == "SQLServer":
                unique_select_statement.append(
                    f'(SELECT TOP {UNIQUE_SAMPLE_LIMIT} STRING_AGG("{col_name}", ',
                    ")"
                    f'WITHIN GROUP (ORDER BY "{col_name}") FROM '
                    f'(SELECT DISTINCT "{col_name}" FROM "{table_name}") sub)'
                    f'AS "unique_values_{col_name}"',
                )

            unique_select_clause = ", ".join(unique_select_statement)
            sql_query = f'SELECT {unique_select_clause} FROM "{table_name}" {where_clause}'
            query = client.sql_query(unique_select_clause, connection=connection)

            unique_values_list = []
            for row in query.iter_rows():
                if row:
                    value = row[0]
                    if value is not None and value != "":
                        unique_values_list.append(str(value))

            final_results[col_name]["unique_values"] = unique_values_list

    if verbose:
        print(f"[METRICS] compute_statistics_sql {round(time.time() - start_time, 3)} seconds")
    return final_results


def compute_statistics_dss(
    project,
    dataset_name,
    numerical_cols,
    categorical_cols,
    aggregations,
    date_col_name,
    date_from=None,
    date_to=None,
    selected_filters: Optional[List[Dict]] = None
):
    dataset = project.get_dataset(dataset_name)
    settings = dataset.get_settings()
    metrics = settings.get_raw()["metrics"]

    # Add the metrics for each column based on its type
    for col in numerical_cols:
        add_metrics_probes(metrics["probes"], "COUNT", col["name"], "col_stats")
        for agg in aggregations:
            if agg.lower() in ["min", "max", "mean", "stddev"]:
                add_metrics_probes(
                    metrics["probes"], agg.upper(), col["name"], "col_stats"
                )
            if agg.lower() == "median":
                add_metrics_probes(
                    metrics["probes"], "P50", col["name"], "percentile_stats"
                )
            if agg.lower() == "q1":
                add_metrics_probes(
                    metrics["probes"], "P25", col["name"], "percentile_stats"
                )
            if agg.lower() == "q3":
                add_metrics_probes(
                    metrics["probes"], "P75", col["name"], "percentile_stats"
                )

    for col in categorical_cols:
        add_metrics_probes(metrics["probes"], "COUNT", col["name"], "col_stats")
        add_metrics_probes(metrics["probes"], "MODE", col["name"], "adv_col_stats", 10)
        add_metrics_probes(
            metrics["probes"], "COUNT_DISTINCT", col["name"], "col_stats"
        )

    # Save settings
    settings.save()
    # TODO: Modify this to load a dataframe with 1 column here! (use get_columns_data)
    # SO Actually, I have to rebuild this whole function...

    # Load the requested column

    # Adjust the prepare recipe based on the date filters
    recipe_name = "compute_product_database_final"
    recipe_settings, _ = get_recipe_settings_and_dictionary(project, recipe_name)
    recipe_payload = recipe_settings.get_json_payload()
    recipe_payload_steps = recipe_payload["steps"]

    # Remove any existing 'FilterOnDate' operation
    original_step_count = len(recipe_payload_steps)
    recipe_payload_steps = [
        step
        for step in recipe_payload_steps
        if "type" in step and step["type"] != "FilterOnDate"
    ]

    filter_removed = original_step_count != len(recipe_payload_steps)

    # If date filters are provided, add a new 'FilterOnDate' operation
    if date_from and date_to:
        date_to_datetime = datetime.strptime(date_to, "%Y/%m/%d")
        date_to_inclusive = (date_to_datetime + timedelta(days=1)).strftime("%Y-%m-%d")

        new_operation = {
            "preview": True,
            "metaType": "PROCESSOR",
            "disabled": False,
            "type": "FilterOnDate",
            "params": {
                "max": f"{date_to_inclusive}",
                "columns": [f"{date_col_name}"],
                "part": "YEAR",
                "values": [],
                "relativeMin": 1,
                "appliesTo": "SINGLE_COLUMN",
                "min": f"{date_from.replace('/', '-')}",
                "booleanMode": "AND",
                "relativeMax": 1,
                "timezone_id": "UTC",
                "action": "KEEP_ROW",
                "filterType": "RANGE",
                "option": "THIS",
            },
            "alwaysShowComment": False,
        }

        recipe_payload_steps.append(new_operation)

    recipe_payload["steps"] = recipe_payload_steps
    recipe_settings.set_json_payload(recipe_payload)
    recipe_settings.save()

    update_recipe_ouput_schema(project, recipe_name)

    if date_from and date_to or filter_removed:
        # Run the scenario
        scenario = project.get_scenario("4REBUILDFINALDATASET")
        scenario.run_and_wait()

    # Retrieve the computed metrics
    dataset.compute_metrics()
    metrics = dataset.get_last_metric_values()
    statistics = {}

    # Process and collect the metrics for each column
    statistics = {}
    # Using defaultdict to initialize nested dictionaries automatically
    col_statistics = defaultdict(dict)

    def get_metric_value(metric_id):
        # Centralize fetching metric value to avoid repetition
        return (
            metrics.get_metric_by_id(metric_id).get("lastValues", [{}])[0].get("value")
        )

    # Process numerical columns
    for col in numerical_cols:
        col_name = col["name"]
        # include count for non null values
        nb_points = get_metric_value(f"col_stats:COUNT:{col_name}")
        if int(nb_points) != 0:
            for agg in aggregations:
                agg_upper = (
                    agg.upper()
                )  # Convert to uppercase as metric IDs are likely in uppercase
                if agg_upper not in ["MEDIAN", "Q1", "Q3"]:
                    metric_id = f"col_stats:{agg_upper}:{col_name}"
                if agg_upper == "MEDIAN":
                    metric_id = f"percentile_stats:P50:{col_name}"
                if agg_upper == "Q1":
                    metric_id = f"percentile_stats:P25:{col_name}"
                if agg_upper == "Q3":
                    metric_id = f"percentile_stats:P75:{col_name}"

                # Directly assign the value to the respective aggregation in col_statistics
                col_statistics[col_name][agg] = get_metric_value(metric_id)

                # Handling median and quartiles differently as dataiku is not returning a list of [min,q1,median,q3,max]
                # It has been fixed so I need to handle both case (returning a list or not)
                if agg_upper in ["MEDIAN", "Q1", "Q3"]:
                    value = col_statistics[col_name][agg]
                    if isinstance(value, str) and value.startswith('[') and value.endswith(']'):
                        # It looks like a list: Remove the brackets and split the string by commas
                        elements = value[1:-1].split(',')
                        # Trim any whitespace and convert to appropriate type
                        elements = [float(element.strip()) for element in elements]
                        # Assign the appropriate value based on agg_upper
                        if agg_upper == "MEDIAN":
                            col_statistics[col_name][agg] = elements[2]
                        elif agg_upper == "Q1":
                            col_statistics[col_name][agg] = elements[1]
                        elif agg_upper == "Q3":
                            col_statistics[col_name][agg] = elements[3]

        col_statistics[col_name]["nb_points"] = nb_points

    # Process categorical columns
    for col in categorical_cols:
        col_name = col["name"]
        nb_points = get_metric_value(f"col_stats:COUNT:{col_name}")
        df = dataiku.Dataset(dataset_name).get_dataframe(columns=[col_name], infer_with_pandas=True)
        statistics[col_name] = {
            "mode": get_metric_value(f"adv_col_stats:MODE:{col_name}"),
            "count_distinct": get_metric_value(f"col_stats:COUNT_DISTINCT:{col_name}"),
            "unique_values": list(df[col_name].astype(str).unique()),
            "nb_points": nb_points,
        }

    # Merge numerical and categorical stats
    statistics.update(col_statistics)

    return statistics


def get_columns_data(
    main_connection: str,
    dataset_name: str,
    date_col_name: str,
    col_list: List[str],
    date_from: Optional[str] = None,
    date_to: Optional[str] = None,
    selected_filters: Optional[List[Dict]] = None,
):
    """
    Fetches data for specified columns within a given date range and optional filter conditions.
    This function is optimized for SQL databases and also supports filesystem-based datasets in Dataiku DSS.

    Parameters:
    - main_connection (str): The name of the database connection.
    - dataset_name (str): The name of the dataset.
    - date_col_name (str): The name of the date column for filtering the data.
    - col_list (List[str]): List of column names to fetch.
    - date_from (Optional[str]): Start date for filtering (inclusive).
    - date_to (Optional[str]): End date for filtering (exclusive).
    - selected_filters (Optional[List[Dict]]): List of filters with target variables and their filtering conditions.
    """
    start_time = time.time()
    # Ensure date_col_name is at the start of col_list if not already included
    if date_col_name not in col_list:
        col_list.insert(0, date_col_name)

    if date_from and date_to:
        date_obj = datetime.strptime(date_to, '%Y-%m-%d')
        date_obj_plus_one_day = date_obj + timedelta(days=1)
        date_to = date_obj_plus_one_day.strftime('%Y-%m-%d')

    client = dataiku.api_client()
    connection_type = client.get_connection(main_connection).get_info()["type"]

    if connection_type in supported_connections:
        dataset = dataiku.Dataset(dataset_name)
        location_info = dataset.get_location_info()
        table_name = location_info.get("info", {}).get("table")

        select_clause = ", ".join([f'"{col}"' for col in col_list])
        where_conditions = []

        if date_from and date_to:
            date_condition = build_date_condition(connection_type, date_col_name, date_from, date_to)
            where_conditions.append(f"{date_condition}")

        if selected_filters:
            filter_condition = build_filter_condition(selected_filters)
            if filter_condition:
                where_conditions.append(filter_condition)

        where_clause = " AND ".join(where_conditions) if where_conditions else "1=1"
        sql_query = f'SELECT {select_clause} FROM "{table_name}" WHERE {where_clause}'
        # Execute the query
        executor = SQLExecutor2(dataset=dataset_name)
        full_dataframe = executor.query_to_df(query=sql_query)

    else:
        # Retrieve the dataset
        dataset = Dataset(dataset_name)
        # Start constructing the dataset filter
        conditions = []

        # Add date range condition
        if date_from and date_to:
            conditions.append(f"({date_col_name} >= '{date_from}')")
            conditions.append(f"({date_col_name} < '{date_to}')")  # '<' ensures up to the end of date_to

        # Add selected filters
        if selected_filters:
            for filter in selected_filters:
                target = filter['target']
                filter_type = filter['filters']['type']
                filter_range = filter['filters']['range']

                if filter_type == "num" and len(filter_range) == 2:
                    conditions.append(f'({target} >= {filter_range[0]})')
                    conditions.append(f'({target} <= {filter_range[1]})')
                elif filter_type == "cat":
                    filter_conditions = [f'{target} == \'{value}\'' for value in filter_range]
                    conditions.append(f'({" || ".join(filter_conditions)})')

        # Combine all conditions
        dataset_filter = ' && '.join(conditions)

        # Fetch the data with the filters applied
        chunks = []
        col_list.insert(0, date_col_name)
        for df_chunk in dataset.iter_dataframes(columns=col_list, filter=dataset_filter, parse_dates=False, chunksize=100_000):
            chunks.append(df_chunk)

        # Concatenate all chunks into a single DataFrame
        full_dataframe = pd.concat(chunks, ignore_index=True)

    print(f"[METRICS] get_columns_data took {round(time.time() - start_time, 3)} seconds")
    return full_dataframe


def compute_counts(
    main_connection,
    dataset_name,
    date_col_name,
    target_variable,
    target_range,
    selected_filters: Optional[List[Dict]] = None,
    date_from: Optional[str] = None,
    date_to: Optional[str] = None,
):

    # Get the connection type
    client = dataiku.api_client()
    connection_type = client.get_connection(main_connection).get_info()["type"]

    if connection_type in supported_connections:
        dataset = dataiku.Dataset(dataset_name)
        location_info = dataset.get_location_info()
        table_name = location_info.get("info", {}).get("table")

        # Initialize the base part of the SQL query
        sql_query = f'SELECT COUNT(CASE WHEN 1=1 THEN 1 END) AS filtered_points, COUNT(CASE WHEN "{target_variable}" IS NULL THEN 1 END) AS null_points'

        # Add target filter to the SQL query
        if target_variable and target_range:
            if isinstance(target_range, list):
                # If the target_range is a list of modalities
                if all(isinstance(item, str) for item in target_range):
                    # For a list of string modalities
                    formatted_values = ", ".join(f"'{value}'" for value in target_range)
                    sql_query += f', COUNT(CASE WHEN "{target_variable}" IN ({formatted_values}) THEN 1 END) as target_filtered_points'
                elif (
                    all(isinstance(item, (int, float)) for item in target_range)
                    and len(target_range) == 2
                ):
                    # For a numerical range with two values
                    sql_query += f', COUNT(CASE WHEN CAST("{target_variable}" AS numeric) BETWEEN {target_range[0]} AND {target_range[1]} THEN 1 END) as target_filtered_points'
                else:
                    raise ValueError(
                        "target_range must be a list of strings or a list of two numbers (int/float)."
                    )
            else:
                raise ValueError("target_range must be a list.")

        sql_query += f' FROM "{table_name}"'

        # Initialize a list to hold different parts of the WHERE clause
        where_conditions = []

        # Add date range condition to the WHERE clause if dates are provided
        if date_from and date_to:
            # Convert and format the date_to to include the entire day
            date_obj = datetime.strptime(date_to, "%Y-%m-%d")
            date_obj_plus_one_day = date_obj + timedelta(days=1)
            date_to = date_obj_plus_one_day.strftime('%Y-%m-%d')
            where_conditions.append(f"\"{date_col_name}\" >= '{date_from}'")
            where_conditions.append(f"\"{date_col_name}\" <= '{date_to}'")

        # Add selected filters to the WHERE clause
        if selected_filters:
            for filter in selected_filters:
                target = filter['target']
                filter_type = filter['filters']['type']
                filter_range = filter['filters']['range']

                if filter_type == "num" and len(filter_range) == 2:
                    where_conditions.append(f'CAST("{target}" AS numeric) BETWEEN {filter_range[0]} AND {filter_range[1]}')
                elif filter_type == "cat":
                    formatted_values = ", ".join(f"'{value}'" for value in filter_range)
                    where_conditions.append(f'\"{target}\" IN ({formatted_values})')

        # Combine all WHERE conditions
        if where_conditions:
            sql_query += " WHERE " + " AND ".join(where_conditions)

        # Execute the query using Dataiku's SQL capabilities
        client = dataiku.api_client()
        query = client.sql_query(
            sql_query, connection=main_connection
        )

        # Fetch the result
        result = next(
            query.iter_rows(), (0, 0, 0)
        )  # Defaults to (0, 0, 0) if no row is returned

    else:
        # Get the target_data
        filtered_target_data = get_columns_data(
            main_connection,
            dataset_name,
            date_col_name,
            [target_variable] + [filter['target'] for filter in selected_filters],
            date_from,
            date_to,
            selected_filters,
        )
        target_filtered_points = len(filtered_target_data[filtered_target_data[target_variable].isin(target_range)])

        target_data = get_columns_data(
            main_connection,
            dataset_name,
            date_col_name,
            [target_variable] + [filter['target'] for filter in selected_filters],
            date_from,
            date_to,
        )
        total_points = len(filtered_target_data)
        null_points = len(target_data[target_data[target_variable].isnull()])

        result = [total_points, null_points, target_filtered_points]

    # Return the total number of points and the number of points that meet the filter criteria
    return {
        "total_points": result[0],
        "filtered_points": result[2],
        "null_points": result[1],
    }


def format_number(num, sig_figs=4):
    if isnan(num):
        return ""

    if num != 0:
        if 0 < abs(num) < 10_000:
            # For numbers between 0 and 10_000, use sig_figs significant figures
            exponent = -int(floor(log10(abs(num)))) + (sig_figs - 1)
            result = round(num, exponent)
            return int(result) if float(result).is_integer() else result
        else:
            # For numbers above 10_000, return the integer part only
            return int(num)
    else:
        return 0  # Handles the case where num is 0


def generate_unique_id(input_string, length=8):
    # Get the current timestamp and concatenate it with the input string
    timestamp = str(time.time()) + input_string

    # Hash the concatenated string using SHA-256 and take the first `length` characters
    hashed_id = hashlib.sha256(timestamp.encode()).hexdigest()[:length]

    return hashed_id


def bin_dataframe(
    df, target_binary, target_column, analysis_variables, input_schema, outliers_filtering, num_bins=10
):
    avg_nok_rates = {}
    df = df[analysis_variables + [target_binary, target_column]].copy()
    output_df = pd.DataFrame(index=df.index)

    for col in df.columns:
        col_type = next(
            (
                column["type"]
                for column in input_schema["columns"]
                if column["name"] == col
            ),
            None,
        )
        pandas_type = df[col].dtype

        if col_type is None:
            if pandas_type == "object":
                col_type = "string"
            elif pandas_type == "int64":
                col_type = "bigint"
            elif pandas_type in ["float64", "float32"]:
                col_type = "double"

        # Check consistency between expected type (from schema) and pandas dtypes
        if col_type in ["double", "bigint"]:
            if pandas_type not in ["float64", "int64"]:
                # Convert column to float or int as needed
                df[col] = pd.to_numeric(df[col], errors="coerce")

        elif col_type == "string" and pandas_type != "object":
            df[col] = df[col].astype(str)

        # Skip if the column is empty
        if len(df[col].dropna().unique()) == 0:
            print(f"Skipping column {col} as it is empty.")
            continue

        try:
            if (
                col_type == "string"
            ):  # If the column is a categorical variable (object type)
                # Calculate the defect rate for categorical data
                nok_points = df[df[target_binary] == "NOK"][col].notna().sum()
                total_valid_points = df[col].notna().sum()
                nok_rate = (nok_points / total_valid_points) * 100

                avg_nok_rates[col] = {
                    "type": "cat",
                    "avg_nok_rate": nok_rate,
                    "total_nb_points": total_valid_points,
                }
                output_df[col] = df[col]

            elif col_type in ["bigint", "double"]:
                # Calculate the defect rate for the column without empty values:
                filtered_df = df.dropna(subset=[col]).copy()
                avg_nok_rate = (filtered_df[target_binary] == "NOK").mean() * 100
                total_nb_points = len(filtered_df[col])

                if outliers_filtering:
                    range_min = np.quantile(filtered_df[col], 0.015)
                    range_max = np.quantile(filtered_df[col], 0.985)
                    filtered_df = filtered_df[
                        (filtered_df[col] >= range_min)
                        & (filtered_df[col] <= range_max)
                    ]
                else:
                    range_min = filtered_df[col].min()
                    range_max = filtered_df[col].max()

                if filtered_df[col].dropna().empty:
                    print(
                        f"Filtered data for column {col} has no valid entries. Skipping binning for this column."
                    )
                    continue

                avg_nok_rates[col] = {
                    "type": "num",
                    "avg_nok_rate": avg_nok_rate,
                    "total_nb_points": total_nb_points,
                }

                # Determine the bins from the filtered data
                _, bins = np.histogram(filtered_df[col].dropna(), bins=num_bins)
                # Cut the dataframe column using the bins
                cut_series = pd.cut(
                    filtered_df[col], bins=bins, include_lowest=True
                )
                output_df.loc[cut_series.index, col] = cut_series

        except AttributeError:
            print(f"Error with column: {col} of type {type(df[col])}")
            raise

    # Add the output_col to the output_df
    output_df[target_binary] = df[target_binary].copy()

    return output_df, avg_nok_rates


def calculate_defect_rate(df, date_col_name, target_binary, target_col, avg_nok_rates):
    impact_list = []
    total_nb_points = len(df[target_binary].dropna())

    for col in df.columns:
        if col not in [date_col_name, target_binary]:
            col_nok_rate = avg_nok_rates[col]["avg_nok_rate"]
            col_nb_points = avg_nok_rates[col]["total_nb_points"]
            variable_type = avg_nok_rates[col]["type"]

            for cat in df[col].dropna().unique():  # Drop NaN values to prevent NaN bins
                slice_df = df[df[col] == cat]
                points_in_bin = len(slice_df)
                bin_nok_rate = (slice_df[target_binary] == "NOK").mean() * 100
                score = np.abs(bin_nok_rate - col_nok_rate) * (
                    points_in_bin / total_nb_points
                )
                color = "green" if bin_nok_rate <= col_nok_rate else "red"
                impact_list.append(
                    {
                        "variable": col,
                        "variable_type": variable_type,
                        "bin": cat,
                        "nok_rate": bin_nok_rate,
                        "points_in_bin": points_in_bin,
                        "impact": score,
                        "color": color,
                        "global_nok_rate": col_nok_rate,
                        "total_nb_points": col_nb_points,
                    }
                )

    impact_df = pd.DataFrame(impact_list)

    variable_impact = impact_df.groupby("variable").agg(
        {"impact": lambda x: np.abs(x).sum(), "global_nok_rate": "last", "variable_type": "last"}
    )

    variable_impact_df = variable_impact.rename(columns={"impact": "relative_score"})

    # Normalize the relative_score
    target_nok_rate = avg_nok_rates['defect_binary']["avg_nok_rate"]
    target_total_nb_points = avg_nok_rates['defect_binary']["total_nb_points"]

    target_nb_points_nok = target_total_nb_points * (target_nok_rate/100)
    target_nb_points_ok = target_total_nb_points - target_nb_points_nok

    max_score = (target_nok_rate) * (target_nb_points_ok/target_total_nb_points) + (100 - target_nok_rate)*(target_nb_points_nok/target_total_nb_points)

    variable_impact_df["relative_score"] = (
        variable_impact_df["relative_score"] / max_score
    ) * 100

    # Dropping target_binary from impact_df and variable_impact_df:
    impact_df = impact_df.loc[impact_df["variable"] != target_binary]
    variable_impact_df = variable_impact_df.loc[variable_impact_df.index != target_binary]

    # Dropping target_col from impact_df and variable_impact_df:
    impact_df = impact_df.loc[impact_df["variable"] != target_col]
    variable_impact_df = variable_impact_df.loc[variable_impact_df.index != target_col]

    # Computing rank:
    variable_impact_df["rank"] = variable_impact_df["relative_score"].rank(
        ascending=False, method="min"
    )

    variable_impact_df = variable_impact_df.sort_values("rank")

    return impact_df, variable_impact_df, total_nb_points


def calculate_zone_impact(variable_df, global_nok_rate):
    """
    Calculate zone impact based on the condition if the defect_rate is below the Average NOK Rate.
    Zones are groups of adjacent bins with the same color.
    """
    most_impacting_green = None
    prev_color = (
        "green" if variable_df["nok_rate"].values[0] <= global_nok_rate else "red"
    )
    start_idx = 0
    max_impact_green = 0

    for idx in range(1, len(variable_df)):
        current_color = (
            "green" if variable_df["nok_rate"].values[idx] <= global_nok_rate else "red"
        )
        if current_color != prev_color:
            impact_sum = variable_df["impact"].values[start_idx:idx].sum()
            if prev_color == "green":
                if impact_sum > max_impact_green:
                    max_impact_green = impact_sum
                    most_impacting_green = (start_idx, idx - 1, prev_color)

            start_idx = idx
            prev_color = current_color

    # Check the last zone
    impact_sum = variable_df["impact"].values[start_idx:].sum()
    if prev_color == "green":
        if impact_sum > max_impact_green:
            most_impacting_green = (start_idx, len(variable_df) - 1, prev_color)

    most_impacting_zones = []
    if most_impacting_green and most_impacting_green[1] - most_impacting_green[0] >= 1:
        return most_impacting_green

    return most_impacting_zones


def calculate_nok_rate(main_connection, connection_type, table_name, target_binary):
    """
    Function to calculate the NOK rate for the given table and target_binary column.
    """
    sql_rate_query = f"""
    SELECT 100 * (
        COUNT(CASE WHEN "{target_binary}" = 'NOK' THEN 1 END) * 1.0 /
        (COUNT(CASE WHEN "{target_binary}" = 'NOK' THEN 1 END) + COUNT(CASE WHEN "{target_binary}" = 'OK' THEN 1 END))
    ) AS NOK_RATE
    FROM "{table_name}";
    """
    try:
        executor = SQLExecutor2(connection=main_connection)
        df = executor.query_to_df(sql_rate_query)
        if connection_type == "PostgreSQL":
            nok_rate = df.iloc[0]['nok_rate']
        elif connection_type == "Snowflake":
            nok_rate = df.iloc[0]['NOK_RATE']
        else:
            raise ValueError(f"Unsupported connection type: {connection_type}")
        return nok_rate
    except Exception as e:
        print(f"An error occurred while calculating NOK rate: {str(e)}")
        return None


def build_date_condition(connection_type, date_col_name, date_from, date_to):
    # Build condition based on connection type
    if connection_type == "PostgreSQL":
        return f"\"{date_col_name}\"::date BETWEEN '{date_from}'::date AND '{date_to}'::date"
    elif connection_type == "SQLServer":
        return f"\"{date_col_name}\" BETWEEN CONVERT(date, '{date_from}') AND CONVERT(date, '{date_to}')"
    elif connection_type == "Snowflake":
        return f"\"{date_col_name}\"::date BETWEEN TO_DATE('{date_from}', 'YYYY-MM-DD') AND TO_DATE('{date_to}', 'YYYY-MM-DD')"
    else:
        raise ValueError(f"Unsupported connection type: {connection_type}")


def build_filter_condition(selected_filters):
    conditions = []
    for filter in selected_filters:
        target = filter['target']
        filter_type = filter['filters']['type']
        filter_range = filter['filters']['range']

        if filter_type == "num":
            conditions.append(f"\"{target}\" BETWEEN {filter_range[0]} AND {filter_range[1]}")
        elif filter_type == "cat":
            formatted_values = ", ".join(f"'{value}'" for value in filter_range)
            conditions.append(f"\"{target}\" IN ({formatted_values})")
    return " AND ".join(conditions)


executed_operations = {}


def handle_target_sql(main_connection,
                      connection_type,
                      table_name,
                      target_binary,
                      target_col,
                      good_range,
                      date_col_name,
                      selected_filters: Optional[List[Dict]] = None,
                      date_from: Optional[str] = None,
                      date_to: Optional[str] = None):

    target_handling_start_time = time.time()

    # Generate a unique key for caching or tracking the operation
    filters_tuples = [
        (filter_dict['target'], tuple(sorted({k: tuple(v) if isinstance(v, list) else v for k, v in filter_dict['filters'].items()}.items())))
        for filter_dict in selected_filters
    ] if selected_filters else []

    key = (target_col['name'], tuple(good_range), tuple(filters_tuples), date_from, date_to)

    # Construct SQL query
    sql_query = f"""UPDATE "{table_name}"
                    SET "{target_binary}" = (
                    CASE WHEN "{target_col['name']}" IS NULL THEN NULL"""

    # Add the 'OK' condition
    if target_col['type'] in numerical_types:
        sql_query += f""" WHEN "{target_col['name']}" BETWEEN {good_range[0]} AND {good_range[1]}"""
    else:
        formatted_values = ", ".join(f"'{value}'" for value in good_range)
        sql_query += f""" WHEN "{target_col['name']}" IN ({formatted_values})"""

    # Add date condition
    if date_from and date_to:
        date_condition = build_date_condition(connection_type, date_col_name, date_from, date_to)
        sql_query += f" AND {date_condition}"

    # Add filter conditions
    if selected_filters:
        filter_condition = build_filter_condition(selected_filters)
        if filter_condition:
            sql_query += f" AND {filter_condition}"

    sql_query += " THEN 'OK'"

    # Add 'NOK' condition
    if target_col['type'] in numerical_types:
        sql_query += f""" WHEN "{target_col['name']}" IS NOT NULL AND ("{target_col['name']}" < {good_range[0]} OR "{target_col['name']}" > {good_range[1]})"""
    else:
        sql_query += f""" WHEN "{target_col['name']}" IS NOT NULL AND "{target_col['name']}" NOT IN ({formatted_values})"""

    # Add repeated conditions for date and filters
    if date_from and date_to:
        sql_query += f" AND {date_condition}"
    if selected_filters:
        sql_query += f" AND {filter_condition}"

    sql_query += " THEN 'NOK' ELSE NULL END)"
    # Execute the SQL query
    try:
        print(f"Handle target sql query:{sql_query}")
        executor = SQLExecutor2(connection=main_connection)
        executor.query_to_df(sql_query, post_queries=["COMMIT"])
    except Exception as e:
        print(f"An error occurred: {str(e)}")
        raise

    if key in executed_operations:
        nok_rate = executed_operations[key]
        print(f"Target already updated for these parameters: {key} and a NOK rate = {nok_rate}. Skipping recalculation...")
        return nok_rate
    else:
        # Calculate NOK rate using the new function
        nok_rate = calculate_nok_rate(main_connection, connection_type, table_name, target_binary)
        if nok_rate is not None:
            executed_operations[key] = nok_rate

        target_handling_end_time = time.time()
        print(f"[METRICS] Target handling done in {(target_handling_end_time - target_handling_start_time):.2f}s, NOK rate={round(nok_rate, 2)}%")

    return nok_rate


def compute_variables_impact_sql(main_connection,
                                 connection_type,
                                 table_name,
                                 input_schema,
                                 global_nok_rate,
                                 target_variable,
                                 analysis_target_variables,
                                 num_bins,
                                 outliers_filtering,
                                 date_col_name: Optional[str] = None,
                                 date_from: Optional[str] = None,
                                 date_to: Optional[str] = None,
                                 selected_filters: Optional[dict] = None,
                                 ):

    requested_variables = int(len(analysis_target_variables))

    # Limiting the number of concurrent workers:
    if requested_variables > 80:
        requested_variables = 80

    # Using ThreadPoolExecutor to execute queries in parallel
    # NOTE: Don't use ProcessPoolExecutor
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=requested_variables
    ) as executor:
        print(f"Number of workers launched: {executor._max_workers}")
        # Future to variable mapping
        future_to_variable = {
            executor.submit(
                compute_variable_result_sql,
                **{
                    'variable': var,
                    'num_bins': num_bins,
                    'table_name': table_name,
                    'main_connection': main_connection,
                    'connection_type': connection_type,
                    'input_schema': input_schema,
                    'outliers_filtering': outliers_filtering,
                    **({'date_col_name': date_col_name} if date_col_name is not None else {}),
                    **({'date_from': date_from} if date_from is not None else {}),
                    **({'date_to': date_to} if date_to is not None else {}),
                    **({'selected_filters': selected_filters} if selected_filters is not None else {})
                }
            ): var
            for var in analysis_target_variables
        }

        # Collecting results in a dictionary
        dataframes = {}
        for future in concurrent.futures.as_completed(future_to_variable):
            variable = future_to_variable[future]
            try:
                data = future.result()
                dataframes[variable] = data
            except Exception as e:
                print(f"Error fetching data for {variable}: {e}")

    impact_list = []
    avg_nok_rates = {}

    for col, df in dataframes.items():

        total_points_in_col = int(pd.to_numeric(df["points_in_bin"]).sum())

        avg_nok_rates[col] = {
            "avg_nok_rate": global_nok_rate,
            "total_nb_points": total_points_in_col,
        }

        variable_type = next((item["type"] for item in input_schema["columns"] if item["name"] == col))

        for _, row in df.iterrows():
            points_in_bin = int(row["points_in_bin"])
            bin_nok_rate = (
                (int(row["nok_count"]) / points_in_bin) * 100 if points_in_bin != 0 else 0
            )
            score = np.abs(bin_nok_rate - global_nok_rate) * (
                points_in_bin / total_points_in_col
            )
            color = "green" if bin_nok_rate <= global_nok_rate else "red"

            # Convert bin to pd.Interval if variable type is numerical
            if variable_type in numerical_types:
                bin_value = parse_interval(row["bin"])
                var_type = "num"

            elif variable_type == "string" or variable_type == "boolean":
                bin_value = row["bin"]
                var_type = "cat"

            impact_list.append(
                {
                    "variable": col,
                    "variable_type": var_type,
                    "bin": bin_value,
                    "nok_rate": bin_nok_rate,
                    "points_in_bin": points_in_bin,
                    "impact": score,
                    "color": color,
                    "global_nok_rate": global_nok_rate,
                    "total_nb_points": total_points_in_col,
                }
            )

    impact_df = pd.DataFrame(impact_list)

    variable_impact = impact_df.groupby("variable").agg(
        {
            "impact": lambda x: np.abs(x).sum(),
            "global_nok_rate": "last",
            "variable_type": "last",
        }
    )

    variable_impact = variable_impact.rename(columns={"impact": "relative_score"})

    # Normalize the relative_score
    target_nok_rate = avg_nok_rates['defect_binary']["avg_nok_rate"]
    target_total_nb_points = avg_nok_rates['defect_binary']["total_nb_points"]

    target_nb_points_nok = target_total_nb_points * (target_nok_rate/100)
    target_nb_points_ok = target_total_nb_points - target_nb_points_nok

    max_score = (target_nok_rate) * (target_nb_points_ok/target_total_nb_points) + (100 - target_nok_rate)*(target_nb_points_nok/target_total_nb_points)

    variable_impact["relative_score"] = (
        variable_impact["relative_score"] / max_score
    ) * 100

    # Dropping target_col from impact_df and variable_impact_df:
    impact_df = impact_df.loc[impact_df["variable"] != target_variable]
    variable_impact = variable_impact.loc[variable_impact.index != target_variable]

    impact_df = impact_df.loc[impact_df["variable"] != "defect_binary"]
    variable_impact = variable_impact.loc[variable_impact.index != "defect_binary"]

    variable_impact["rank"] = variable_impact["relative_score"].rank(
        ascending=False, method="min"
    )
    variable_impact = variable_impact.sort_values("rank")

    return impact_df, variable_impact


def compute_numerical_bins_sql(main_connection,
                               connection_type,
                               table_name,
                               selected_numerical_variables,
                               outliers_filtering,
                               date_col_name,
                               num_bins=10,
                               date_from=None,
                               date_to=None,
                               selected_filters=None):
    # Prepare SQL query parts for calculating range_min and range_max
    query_parts = []
    for variable in selected_numerical_variables:
        if outliers_filtering:
            query_parts.append(
                f"""
                PERCENTILE_CONT(0.015) WITHIN GROUP (ORDER BY "{variable}") AS "range_min_{variable.lower()}",
                PERCENTILE_CONT(0.985) WITHIN GROUP (ORDER BY "{variable}") AS "range_max_{variable.lower()}"
                """
            )
        else:
            query_parts.append(
                f"""
                MIN("{variable}") AS "range_min_{variable.lower()}",
                MAX("{variable}") AS "range_max_{variable.lower()}"
                """
            )

    # Joining all percentile calculations into one query
    complete_query = "SELECT\n" + ",\n".join(query_parts) + f'\nFROM "{table_name}"'

    # Prepare SQL query parts for filtering using build_filter_condition function
    conditions = []
    if selected_filters:
        filter_condition = build_filter_condition(selected_filters)
        if filter_condition:
            conditions.append(filter_condition)

    if date_from and date_to:
        date_condition = build_date_condition(connection_type, date_col_name, date_from, date_to)
        if date_condition:
            conditions.append(date_condition)

    where_clause = " WHERE " + " AND ".join(conditions) if conditions else ";"

    complete_query += where_clause + ';'

    # Execute the query
    client = dataiku.api_client()
    query = client.sql_query(complete_query, connection=main_connection)
    query.verify()

    # Retrieve results
    results = []
    for row in query.iter_rows():
        results.append(row)

    # Assuming one row of result
    column_names = [
        f"range_min_{var.lower()}" for var in selected_numerical_variables
    ] + [f"range_max_{var.lower()}" for var in selected_numerical_variables]
    df = pd.DataFrame([results[0]], columns=column_names)

    # Data Dictionary for range min and max
    data = []
    for variable in selected_numerical_variables:
        data.append(
            {
                "Variable Name": variable,
                "range_min": df[f"range_min_{variable.lower()}"].iloc[0],
                "range_max": df[f"range_max_{variable.lower()}"].iloc[0],
            }
        )

    df_final = pd.DataFrame(data)

    # Generating bin edges and intervals for each variable
    variables_bins = {}
    for _, row in df_final.iterrows():
        variable = row["Variable Name"]
        range_min = float(row["range_min"])
        range_max = float(row["range_max"])

        bin_edges = np.linspace(range_min, range_max, num=num_bins + 1, endpoint=True)
        intervals = [(bin_edges[i], bin_edges[i + 1]) for i in range(num_bins)]

        variables_bins[variable] = intervals

    return variables_bins


def compute_categorical_bins_sql(
        main_connection,
        connection_type,
        table_name,
        selected_categorical_variables,
        date_col_name=None,
        date_from=None,
        date_to=None,
        selected_filters=None
        ):

    client = dataiku.api_client()

    complete_categorical_query = generate_top_modalities_query_with_filters(connection_type,
                                                                            table_name,
                                                                            selected_categorical_variables,
                                                                            date_col_name,
                                                                            date_from,
                                                                            date_to,
                                                                            selected_filters)

    try:
        # Execute the query
        query = client.sql_query(complete_categorical_query, connection=main_connection)
        query.verify()

        # Retrieve results
        results = []
        for row in query.iter_rows():
            results.append(row)

        # Create DataFrame with required columns
        output_df = pd.DataFrame(
            results, columns=["variable", "bin", "points_in_bin", "nok_count"]
        )
        # Ensure correct data types
        output_df['bin'] = output_df['bin'].astype(str)
        output_df['points_in_bin'] = output_df['points_in_bin'].astype(int)
        # Filter out rows where "bin" is "Others" and "points_in_bin" is 0
        output_df = output_df[((output_df['bin'] != 'Others') & (output_df['points_in_bin'] != 0))]

    except Exception as e:
        print(
            f"Error while executing query for table '{table_name}':\n{complete_categorical_query}\nException: {e}"
        )
        raise

    return output_df


def generate_top_modalities_query_with_filters(connection_type,
                                               table_name,
                                               selected_categorical_variables,
                                               date_col_name,
                                               date_from=None,
                                               date_to=None,
                                               selected_filters=None,
                                               top_n=30):

    # Prepare SQL query parts for filtering using build_filter_condition function
    conditions = []
    if selected_filters:
        filter_condition = build_filter_condition(selected_filters)
        if filter_condition:
            conditions.append(filter_condition)

    if date_from and date_to:
        date_condition = build_date_condition(connection_type, date_col_name, date_from, date_to)
        if date_condition:
            conditions.append(date_condition)

    where_clause = " WHERE " + " AND ".join(conditions) if conditions else ""

    categorical_queries = []
    for variable in selected_categorical_variables:
        query = f"""
        WITH TopModalities AS (
            SELECT
                "{variable}" AS modality,
                COUNT("defect_binary") AS point_count
            FROM
                "{table_name}"
            {where_clause}
            GROUP BY
                "{variable}"
            ORDER BY
                point_count DESC
            LIMIT {top_n}
        )
        SELECT
            '{variable}' AS variable,
            tm.modality AS bin,
            COUNT("defect_binary") AS points_in_bin,
            SUM(CASE WHEN "defect_binary" = 'NOK' THEN 1 ELSE 0 END) AS nok_count
        FROM
            "{table_name}" t
        JOIN TopModalities tm ON t."{variable}" = tm.modality
        {where_clause}
        GROUP BY
            tm.modality
        ORDER BY
            tm.modality;
        """
        categorical_queries.append(query)

    # Combine all queries into one
    complete_categorical_query = "\nUNION ALL\n".join(categorical_queries)

    return complete_categorical_query


def generate_bin_query(variables_bins, table_name):
    case_statements_bin = []

    # Assuming 'variables_bins' has only one variable and its bins
    variable, bins = next(iter(variables_bins.items()))

    for bin_range in bins:
        lower_bound, upper_bound = bin_range
        case_statements_bin.append(
            f"""WHEN "{variable}" BETWEEN {lower_bound} AND {upper_bound} THEN '({lower_bound}, {upper_bound}]'"""
        )

    bin_case_statement = "CASE " + " ".join(case_statements_bin) + " ELSE NULL END AS bin"

    query = f"""
    SELECT
        bin,
        COUNT("defect_binary") AS points_in_bin,
        SUM(CASE WHEN "defect_binary" = 'NOK' THEN 1 ELSE 0 END) AS nok_count
    FROM (
        SELECT
            {bin_case_statement},
            "defect_binary",
            "{variable}" AS \"{variable}\"
        FROM "{table_name}"
    ) AS derived
    WHERE \"{variable}\" IS NOT NULL
    GROUP BY bin;
    """
    return query


logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)


def compute_variable_result_sql(variable,
                                num_bins,
                                table_name,
                                main_connection,
                                connection_type,
                                input_schema,
                                outliers_filtering,
                                date_col_name: Optional[str] = None,
                                date_from: Optional[str] = None,
                                date_to: Optional[str] = None,
                                selected_filters: Optional[str] = None):
    variable_result_start_time = time.time()

    # Mapping variable names to their types in input_schema
    variable_types = {col["name"]: col["type"] for col in input_schema["columns"]}

    is_numerical = variable_types.get(variable) in numerical_types
    is_categorical = variable_types.get(variable) == "string" or "boolean"

    try:
        if is_numerical:
            # Compute bins dynamically for each variable
            variables_bins = compute_numerical_bins_sql(
                main_connection, connection_type, table_name, [variable], outliers_filtering, date_col_name, num_bins, date_from, date_to, selected_filters)
            # Generate the SQL query using the computed bins
            bin_query = generate_bin_query(
                {variable: variables_bins[variable]}, table_name
            )

            # Execute the query using a fresh instance of Dataiku's API client for each variable
            client = dataiku.api_client()  # Create a new client instance for each function call
            print(f"for variable {variable}, bin query is:{bin_query}")
            query = client.sql_query(bin_query, connection=main_connection)
            query.verify()

            # Fetch the results and load into a DataFrame
            data = pd.DataFrame(
                query.iter_rows(),
                columns=["bin", "points_in_bin", "nok_count"],
            )

            # insert the variable column
            data.insert(0, 'variable', variable)
            data.name = f"{variable}_data"

            # Filter out rows where 'bin' is empty or NULL
            data = data[data["bin"].notnull() & (data["bin"] != '')]

            logger.info(f"Data successfully retrieved for num var {variable} in {round(time.time() - variable_result_start_time, 2)}s, rows: {len(data)}")
            return data

        elif is_categorical:
            data = compute_categorical_bins_sql(
                main_connection, connection_type, table_name, [variable], date_col_name, date_from, date_to, selected_filters)
            logger.info(f"Data successfully retrieved for cat var {variable} in {round(time.time() - variable_result_start_time, 2)}s, rows: {len(data)}")
            return data

    except Exception as e:
        logger.error(f"Error processing {variable}: {str(e)}")
        traceback.print_exc()


def parse_interval(bin_value):
    # Determine the type of closure based on the brackets
    if bin_value[0] == "[":
        left_closed = True
    else:
        left_closed = False

    if bin_value[-1] == "]":
        right_closed = True
    else:
        right_closed = False

    # Extract the numeric values
    numbers = bin_value.strip("[]()").split(",")
    low = float(numbers[0].strip())
    high = float(numbers[1].strip())

    # Create the interval
    if left_closed and right_closed:
        closure = "both"
    elif left_closed and not right_closed:
        closure = "left"
    elif not left_closed and right_closed:
        closure = "right"
    else:
        closure = "neither"

    return pd.Interval(low, high, closed=closure)


def compute_current_state_sql(
    main_connection,
    dataset_name,
    date_col_name,
    filter_criteria,
    target_binary,
    date_from: Optional[str] = None,
    date_to: Optional[str] = None,
):
    # Get the connection type
    client = dataiku.api_client()
    connection_type = client.get_connection(main_connection).get_info()["type"]

    if connection_type in supported_connections:
        dataset = dataiku.Dataset(dataset_name)
        location_info = dataset.get_location_info()
        table_name = location_info.get("info", {}).get("table")

        # Initialize the base part of the SQL query
        sql_query = f"""
            SELECT
                COUNT("defect_binary") as total_points,
                COUNT(CASE WHEN "{target_binary}" IS NULL THEN 1 END) AS null_count,
                COUNT(CASE WHEN "{target_binary}" = 'NOK' THEN 1 END) AS nok_count
        """

        # Construct WHERE conditions based on filter criteria
        where_conditions = []
        for filter_variable, filter_range in filter_criteria.items():

            if isinstance(filter_range, list):
                if all(isinstance(item, str) for item in filter_range):
                    formatted_values = ", ".join(f"'{value}'" for value in filter_range)
                    where_conditions.append(f""" "{filter_variable}" IN ({formatted_values})""")
                elif all(isinstance(item, (int, float)) for item in filter_range) and len(filter_range) == 2:
                    where_conditions.append(f""" "{filter_variable}" BETWEEN {filter_range[0]} AND {filter_range[1]}""")
                else:
                    raise ValueError("Each filter_range must be either a list of strings or a list of two numbers (int/float).")
            else:
                raise ValueError("filter_range must be a list.")

        sql_query += f' FROM "{table_name}"'

        # Add date range condition to the WHERE clause if dates are provided
        if date_from and date_to:
            build_date_condition(connection_type, date_col_name, date_from, date_to)
            """
            TO VERIFY: I update the code below to the function build_date_condition
            date_obj = datetime.strptime(date_to, "%Y-%m-%d")
            date_obj_plus_one_day = date_obj + timedelta(days=1)
            new_date_to = date_obj_plus_one_day.strftime("%Y-%m-%d")
            where_conditions.append(f"\"{date_col_name}\" >= '{date_from}'")
            where_conditions.append(f"\"{date_col_name}\" < '{new_date_to}'")  # Use < for the upper boundary
            """

        # Combine all WHERE conditions
        if where_conditions:
            sql_query += " WHERE " + " AND ".join(where_conditions)

        logger.info("Generated SQL query: %s", sql_query)

        # Execute the query using Dataiku's SQL capabilities
        query = client.sql_query(
            sql_query, connection=main_connection
        )

        # Fetch the result
        result = next(
            query.iter_rows(), (0, 0, 0)
        )  # Defaults to (0, 0, 0) if no row is returned

        return result
