import { defineStore } from "pinia";
import { API } from "../Api";
import { SavedStudy, Variable } from "../components/types";
import { formatDate, formatDateRange, formatNumbers } from "../utilities/utils";
import { useStore } from "./webapp";

export const useStudiesStore = defineStore("StudieStore", {
    state: () => ({
        studies: [] as SavedStudy[],
        selectedStudy: undefined as undefined | SavedStudy,
    }),
    getters: {
        graphTypeForTab(state) {
            return (index: number) => {
                if (index < TABS_KEYS.length && index >= 0) {
                    return state.graphType[TABS_KEYS[index]];
                } else {
                    console.log("ERROR TAB INDEX IS OUT OF RANGE", index);
                    return "standard";
                }
            };
        },
    },
    actions: {
        selectStudy(study: SavedStudy) {
            this.selectedStudy = study;
        },
        async fetchStudies() {
            try {
                let resp = await API.getStudies();
                // Process the response data
                let studies = resp.studies;
                this.studies = studies
                    .sort((a, b) => new Date(b.date) - new Date(a.date))
                    .map((study: any) => {
                        let newStr = study.date_range.replace(/\//g, "-");
                        const formattedDate = formatDate(
                            study.date
                        ).fullWithUSTimestamp;
                        const dateRange = formatDateRange(JSON.parse(newStr));
                        const dates = JSON.parse(study.date_range);
                        const parsedFilters = JSON.parse(study.selected_filters.replace(/\'/g, '\"'));

                        return {
                            id: study.id,
                            studyDate: formattedDate,
                            userEmail: study.user,
                            target: study.target,
                            dateRange: dateRange,
                            fullDate: { from: dates[0], to: dates[1] },
                            targetType: study.target_type || "num",
                            targetInfo: {
                                //todo review
                                target: study.target,
                                targetDate: {
                                    from: JSON.parse(study.date_range)[0].split(
                                        " "
                                    )[0],
                                    to: JSON.parse(study.date_range)[1].split(
                                        " "
                                    )[0],
                                },
                                targetType: study.target_type || "num",
                                targetSelectedRange: study.target_info.target_selected_range
                            },
                            selectedFilters: parsedFilters,
                        };
                    });
                return { studies: this.studies };
            } catch (err) {
                console.log("error:", err);
                throw new Error("Failed to retrieve studies");
            }
        },
        async checkStudy() {
            if (this.selectedStudy) {
                try {
                    const resp = await API.checkStudy(this.selectedStudy.id);
                    if (resp.missing_variables && resp.missing_variables.length) {
                        this.selectedStudy.selectedVariables = this.selectedStudy.selectedVariables.filter(el => resp.missing_variables.indexOf(el) >= 0);
                        this.selectedStudy.missingVariables = resp.missing_variables;
                    }
                    if (resp.ranges_error) {
                        this.selectedStudy.savedVariables.forEach((el) => {
                            if (resp.ranges_error[el.name]) {
                                el.error = resp.ranges_error[el.name].error;
                            }
                            if(resp.missing_variables.indexOf(el.name) >= 0) {
                                el.error = "Variable not found"
                            }
                        });
                    }
                } catch (err) {
                    console.log("error:", err);
                    throw new Error("Failed to check study's information");
                }
            }
        },
        async fetchStudyDetails() {
            if (this.selectedStudy) {
                try {
                    let resp = await API.getStudy(this.selectedStudy.id);
                    let graphs = resp.graphs;
                    let initalState = {
                        ...resp.initial_state,
                        title: "Initial population",
                    };
                    let currentState = {
                        ...resp.final_state,
                        title: "Final population",
                    };
                    let ratio = {
                        "% Initial population": formatNumbers(
                            resp.ratio.pct_initial_pop
                        ),
                        "% Improvement": formatNumbers(
                            resp.ratio.pct_improvement
                        ),
                    };
                    let targetInfo = resp.target_info
                        ? {
                              target: resp.target_info.target,
                              targetDate: resp.target_info.target_date,
                              targetRange:
                                  resp.target_info.target_selected_range,
                              initialRange:
                                  resp.target_info.target_initial_range,
                              targetType: resp.target_info.target_type,
                              targetSelectedRange: resp.target_info.target_selected_range
                          }
                        : this.selectedStudy.targetInfo;
                    this.selectedStudy.targetInfo = targetInfo;
                    this.selectedStudy.initalState = initalState;
                    this.selectedStudy.currentState = currentState;
                    this.selectedStudy.ratio = ratio;
                    this.selectedStudy.selectedVariables =
                        resp.selected_variables;
                    this.selectedStudy.savedVariables = resp.saved_variables;
                    this.selectedStudy.bin = resp.target_info?.nb_bins;
                    this.selectedStudy.selectedFilters = resp.selected_filters;
                    return {
                        graphs,
                        initalState,
                        currentState,
                        ratio,
                        targetInfo,
                        selectedVariables: resp.selected_variables,
                    };
                } catch (err) {
                    console.log("error:", err);
                    throw new Error("Failed to retrieve study");
                }
            }
        },
        async deleteStudy() {
            if (this.selectedStudy) {
                try {
                    await API.deleteStudy(this.selectedStudy.id);
                    this.studies = this.studies.filter(
                        (study) => study.id !== this.selectedStudy!.id
                    );
                    this.selectedStudy = undefined;
                } catch (err) {
                    throw new Error("Failed to retrieve studies");
                }
            }
        },
        async saveStudy(id: string | undefined, variables: Variable[]) {
            try {
                const targetStore = useStore();
                const resp = API.saveStudy({
                    study_id: id,
                    variables: variables,
                    target_info: {
                        target: targetStore.targetInfo.target,
                        target_date: targetStore.targetInfo.targetDate,
                        target_selected_range:
                            targetStore.targetInfo.targetRange,
                        target_initial_range:
                            targetStore.targetInfo.initialRange,
                        target_type: targetStore.targetInfo.targetType,
                    },
                    analysis_variables: targetStore.selectedAnalysisVars,
                    selectedFilters: await targetStore.getSelectedFilters()
                });
                const studyId = resp.study_id;
                return { studyId };
            } catch (err) {
                console.log("error:", err);
                throw new Error("Failed to save study");
            }
        },
    },
});
