import { defineStore } from "pinia";
import { API } from "../Api";
import { useStudiesStore } from "./studies";

import {
    DateRange,
    Graph,
    State,
    Variable,
    VariableRange,
    Statistics,
} from "../components/types";
import { formatNumbers } from "../utilities/utils";

interface TargetInfo {
    target: string;
    targetDate?: DateRange;
    targetRange?: VariableRange;
    initialRange?: VariableRange;
    targetType?: "num" | "cat";
}
interface DraftTargetInfo extends TargetInfo {
    chartData?: any;
    sortedData?: { green: any; red: any };
    distribution?: { good: any; bad: any };
    selectedRange?: VariableRange;
    submitted: boolean;
}
export interface DraftFilterInfo extends DraftTargetInfo {
    id: string;
    isCategorical: boolean;
    hasValueChanged: boolean;
    newNumRange:VariableRange | { min: number; max: number } | undefined;
    newCatRange: VariableRange | { min: number; max: number } | undefined;

}
export const TABS_KEYS = ["firstTab", "secondTab", "thirdTab", "forthTab"];
export const useStore = defineStore("WebAppStore", {
    state: () => ({
        targetInfo: {} as TargetInfo,
        draftTarget: {} as DraftTargetInfo,
        draftFilters: [] as DraftFilterInfo[],
        graphType: {
            secondTab: "standard",
            thirdTab: "standard",
            forthTab: "standard",
        },
        targetVariables: [] as string[],
        analysisGraphs: [] as Graph[],
        selectedBin: 10,
        selectedAnalysisVars: [] as string[],
        savedVariables: [] as Variable[],
        initialState: undefined as undefined | State,
        currentState: undefined as undefined | State,
        ratio: undefined as undefined | State,
    }),
    getters: {
        graphTypeForTab(state) {
            return (index: number) => {
                if (index < TABS_KEYS.length && index >= 0) {
                    return state.graphType[TABS_KEYS[index]];
                } else {
                    console.log("ERROR TAB INDEX IS OUT OF RANGE", index);
                    return "standard";
                }
            };
        },
    },
    actions: {
        async copyStudy(study: any) {
            if (!study.targetInfo?.target) {
                throw new Error("Failed to copy study due to missing data");
                return;
            }
            try {
                if (
                    !study.targetInfo ||
                    !study.targetInfo.target ||
                    !study.targetInfo.targetDate ||
                    !study.targetInfo.targetRange ||
                    !study.targetInfo.initialRange ||
                    !study.targetInfo.targetType ||
                    !study.selectedVariables ||
                    !study.savedVariables
                ) {
                    throw new Error(
                        "Failed to copy selected study due to missing information"
                    );
                }
                this.draftTarget = {
                    ...study.targetInfo,
                    selectedRange: [...study.targetInfo.targetRange],
                    targetRange: [...study.targetInfo.initialRange],
                };
                const studiesStore = useStudiesStore();
                await studiesStore.checkStudy();
                this.selectedAnalysisVars = study.selectedVariables;
                this.analysisGraphs = [];
                this.selectedBin = study.bin ?? this.selectedBin;
                this.initialState = study.initalState;
                this.currentState = {
                    ...study.currentState,
                    title: "Currently selected population",
                };
                this.ratio = study.ratio;
                this.savedVariables = study.savedVariables;
                await this.createTargetChart();
                this.draftTarget.submitted = true;
                this.targetInfo = { ...study.targetInfo };
                await this.updateTarget();
            } catch (err) {
                throw new Error("Failed to copy study");
            }
        },
        submitTargetInfo() {
            this.draftTarget.submitted = true;
            this.targetInfo = {
                target: this.draftTarget.target,
                targetRange: this.draftTarget.selectedRange,
                targetDate: this.draftTarget.targetDate,
                targetType: this.draftTarget.targetType,
                initialRange: this.draftTarget.targetRange,
            };
            if (this.selectedAnalysisVars.length) {
                this.selectVariables(
                    this.selectedAnalysisVars.filter(
                        (el) => el !== this.draftTarget.target
                    ) || []
                );
            } else {
                this.selectVariables([]);
            }
            this.updateAnalysisGraphs([]);
        },
        selectVariables(vars: string[]) {
            this.selectedAnalysisVars = vars;
        },
        selectRange(newVal: VariableRange | undefined) {
            if (!newVal) {
                this.draftTarget.targetRange = newVal;
            }
            this.draftTarget.selectedRange = newVal;
        },
        selectRangeFilter(newVal: VariableRange | { min: number; max: number } | undefined, index: number) {
            for (let i = 0; i < this.draftFilters.length; i++) {
                if (i === index) {
                    if (!newVal) {
                        this.draftFilters[i].targetRange = newVal;
                    }
                    this.draftFilters[i].selectedRange = newVal;
                }
            }
        },
        saveVariables(vars: Variable[]) {
            this.savedVariables = vars;
            if (!vars.length) {
                // todo check
                this.currentState = undefined;
                this.ratio = undefined;
            }
        },
        updateAnalysisGraphs(graphs: Graph[]) {
            this.analysisGraphs = graphs;
        },
        changeGraphType(index: number, type: string) {
            if (
                index >= 0 &&
                index < TABS_KEYS.length &&
                this.graphType[TABS_KEYS[index]]
            ) {
                this.graphType[TABS_KEYS[index]] = type;
            }
        },
        resetTargetRange() {
            this.draftTarget.targetRange = undefined;
        },
        buildStatistics(stats: Statistics): any {
            if (this.draftTarget.targetType === "num") {
                return [
                    {
                        Max: stats.max ? formatNumbers(stats.max) : 0,
                        Q3: stats.upper_quartile
                            ? formatNumbers(stats.upper_quartile)
                            : 0,
                        Median: stats.median ? formatNumbers(stats.median) : 0,
                        Q1: stats.lower_quartile
                            ? formatNumbers(stats.lower_quartile)
                            : 0,
                        Min: stats.min ? formatNumbers(stats.min) : 0,
                    },
                    {
                        Points: formatNumbers(stats.nbr_points),
                        Mean: stats.mean ? formatNumbers(stats.mean) : 0,
                        "std. dev": stats.std_dev
                            ? formatNumbers(stats.std_dev)
                            : 0,
                    },
                ];
            } else {
                return [{
                    Points: formatNumbers(stats.nbr_points),
                    "Distinct values": formatNumbers(stats.distinct_values),
                    Mode: stats.mode,
                }];
            }
        },
        async fetchTargets() {
            try {
                const resp = await API.getTargetVariables();
                return {
                    targets: resp.targets,
                    dateRange: resp.date_range,
                };
            } catch (err) {
                throw new Error("Failed to retrieve targets");
            }
        },
        async fetchTargetRange() {
            if (this.draftTarget.target) {
                try {
                    const resp = await API.getTargetRange({
                        target: this.draftTarget.target,
                        dates: [
                            this.draftTarget.targetDate?.from,
                            this.draftTarget.targetDate?.to,
                        ],
                        selectedFilters: await this.getSelectedFilters()
                    });
                    if (resp.target[this.draftTarget.target].unique_values) {
                        this.draftTarget.targetRange =
                            resp.target[this.draftTarget.target].unique_values;
                        this.draftTarget.targetType = "cat";
                    } else {
                        this.draftTarget.targetRange = [
                            resp.target[this.draftTarget.target].min,
                            resp.target[this.draftTarget.target].max,
                        ];
                        this.draftTarget.targetType = "num";
                    }
                    return { targetRange: { ...this.draftTarget.targetRange } };
                } catch (err) {
                    throw new Error("Failed to retrieve data");
                }
            }
        },
        async fetchFilterRange(index: number) {
            for (let i = 0; i < this.draftFilters.length; i++) {
                if (i !== index) {
                    continue;
                }

                let filter = this.draftFilters[i];
                if (filter) {
                    try {
                        const resp = await API.getFilterRange({
                            target: filter.target,
                            dates: [
                                this.draftTarget.targetDate?.from,
                                this.draftTarget.targetDate?.to,
                            ],
                        });
                        if (resp.target[filter.target].unique_values) {
                            filter.targetRange = resp.target[filter.target].unique_values;
                            filter.targetType = "cat";
                            filter.isCategorical = true;
                            if (!filter.newCatRange) {
                                filter.newCatRange = [];
                            }
                        } else {
                            filter.targetRange = [
                                resp.target[filter.target].min,
                                resp.target[filter.target].max,
                            ];
                            filter.targetType = "num";
                            filter.isCategorical = false;
                            if (!filter.newNumRange) {
                                filter.newNumRange = { min: undefined, max: undefined };
                            }
                        }
                    } catch (err) {
                        throw new Error("Failed to retrieve data");
                    }
                }
            }
        },
        async getSelectedFilters() {
            let selectedFilters = [];
            for (let o of this.draftFilters) {
                selectedFilters.push({
                    target: o.target,
                    filters: {
                        range: o.selectedRange,
                        type: o.targetType
                    }
                });
            }
            return selectedFilters;
        },
        async createTargetChart() {
            if (this.draftTarget.target) {
                try {
                    this.draftTarget.chartData = undefined;
                    this.draftTarget.distribution = undefined;
                    this.draftTarget.sortedData = undefined;
                    
                    const resp = await API.getTargetChart({
                        target: this.draftTarget.target,
                        dates: [
                            this.draftTarget.targetDate?.from,
                            this.draftTarget.targetDate?.to,
                        ],
                        selectedFilters: await this.getSelectedFilters()
                    });
                    this.draftTarget.chartData = {
                        data: resp.data,
                        error: resp.error_message,
                        statistics: this.buildStatistics(resp.statistics),
                    };
                    return {
                        ...this.draftTarget.chartData,
                    };
                } catch (err) {
                    throw new Error("Failed to retrieve data");
                }
            }
        },
        async updateTarget() {
            if (this.draftTarget.target) {
                try {
                    this.draftTarget.sortedData = undefined;
                    this.draftTarget.distribution = undefined;
                    const resp = await API.updateTarget({
                        target: this.draftTarget.target,
                        good_range: [...this.draftTarget.selectedRange!],
                        selectedFilters: await this.getSelectedFilters()
                    });
                    this.draftTarget.sortedData = {
                        green: resp.green_data,
                        red: resp.red_data,
                    };
                    this.draftTarget.distribution = {
                        good: resp.nb_green_points,
                        bad: resp.nb_red_points,
                    };
                    return {
                        sortedData: { ...this.draftTarget.sortedData },
                        distribution: { ...this.draftTarget.distribution },
                    };
                } catch (err) {
                    throw new Error("Failed to change settings");
                }
            }
        },
        async fetchTargetVariables() {
            if (this.targetInfo.target) {
                try {
                    let resp = await API.getAnalysisVariables(
                        this.targetInfo.target
                    );
                    this.targetVariables = resp.variables;
                    return this.targetVariables;
                } catch (err) {
                    throw new Error("Failed to retrieve targets");
                }
            }
        },
        async getTargetState() {
            if (this.targetInfo.target) {
                try {
                    let resp = await API.getTargetState({
                        target: this.targetInfo.target,
                        filters: {
                            dates: [
                                this.targetInfo.targetDate!.from,
                                this.targetInfo.targetDate!.to,
                            ],
                            range: [...this.targetInfo.targetRange!],
                        },
                        selectedFilters: await this.getSelectedFilters()
                    });
                    this.initialState = resp.initial_state;
                    this.currentState = undefined;
                    this.ratio = undefined;
                    return {
                        variables: resp.variables,
                        initialState: resp.initial_state,
                    };
                } catch (err) {
                    throw new Error("Failed to retrieve data");
                }
            }
        },
        async updateTargetState(
            variable: Variable,
            vars: {
                variable: string;
                good_range: VariableRange;
            }[],
            action: "delete" | "add"
        ) {
            if (this.targetInfo.target) {
                try {
                    let resp = await API.updateTargetState({
                        target: this.targetInfo.target,
                        filters: {
                            dates: [
                                this.targetInfo.targetDate!.from,
                                this.targetInfo.targetDate!.to,
                            ],
                            range: [...this.targetInfo.targetRange!],
                        },
                        variables: vars,
                        selectedFilters: await this.getSelectedFilters()
                    });
                    this.currentState = {
                        ...resp.current_state,
                        title: "Currently selected population",
                    };
                    this.ratio = {
                        "% Initial population": formatNumbers(
                            resp.ratio.pct_initial_pop
                        ),
                        "% Improvement": formatNumbers(
                            resp.ratio.pct_improvement
                        ),
                    };
                    this.savedVariables = this.savedVariables.filter(
                        (el: Variable) => el.name !== variable.name
                    );
                    if (action === "add") {
                        this.savedVariables.push(variable!);
                    }
                    return {
                        savedVariables: this.savedVariables,
                        currentState: this.currentState,
                        ratio: this.ratio,
                    };
                } catch (err) {
                    throw new Error("Failed to save range");
                }
            }
        },
        async getAnalysisGraphs(selectedVariables: string[]) {
            if (this.targetInfo.target) {
                try {
                    let resp = await API.getAnalysis({
                        target: this.targetInfo.target,
                        bin: this.selectedBin,
                        filters: {
                            variables: selectedVariables,
                            good_range: [...this.targetInfo.targetRange!],
                            dates: [
                                this.targetInfo.targetDate!.from,
                                this.targetInfo.targetDate!.to,
                            ],
                        },
                        selectedFilters: await this.getSelectedFilters()
                    });
                    this.analysisGraphs = resp.graphs.sort(
                        (a: Graph, b: Graph) => a.rank - b.rank
                    );
                    if (this.savedVariables) {
                        this.savedVariables.forEach((el) => {
                            const graphData = this.analysisGraphs.find(
                                (graph: Graph) => graph.variable === el.variable
                            );
                        });
                    }
                    return { graphs: this.analysisGraphs };
                } catch (err) {
                    throw new Error("Variables analysis failed");
                }
            }
        },
    },
});
