import { formatDate, formatNumbers } from "./utils.ts";
export enum ChartType {
  TargetScatter,
  AnalysisScatter,
  Pie,
  AnalysisArea
}
function getShape(shape: 'circle' | 'rectangle', color: string): string {
  switch (shape) {
    case 'rectangle':
      return `<svg width="15" height="10"><rect width="15" height="10" style="fill:${color};" opacity="0.7" /></svg>`;
    case 'circle':
      return `<svg width="24" height="12" viewBox="0 0 24 12" fill="none" xmlns="http://www.w3.org/2000/svg">
          <rect y="5" width="24" height="2" fill="${color}"/>
          <circle opacity="0.7" cx="12" cy="6" r="6" fill="${color}"/>
          </svg>
          `;
    default:
      return '';
  }
}
function template(label: string, ...values: { key: string, value: string }[]): string {
  return `
    <div style="display: flex; flex-direction: column; align-items: flex-start; gap: 4px;">
        <div style="border-bottom: 1px solid #ccc; width: 100%; text-align: left; padding-bottom: 4px; color: #666666" class="bs-font-medium-2-semi-bold">
            ${label}
        </div>
        ${values.map(val => `
          <div style="display: flex; justify-content: space-between; align-items: center; width: 100%;gap: 20px;">
              <span class="dku-tiny-text" style="font-size:14px">${val.key}</span>
              <span class="dku-tiny-text" style="font-size:14px"><b>${val.value} </b></span>
          </div>`).join("")}
    </div>
`;
}


export function scatterTooltip(params: any): string {
  const label = formatDate(params.data[0]).fullWithTimestamp;
  const formatted = formatNumbers(params.data[1]);
  const value = formatted == 'NaN' ? params.data[1] : formatted;
  return template(label, { key: params.seriesName, value });
}

export function pieTooltip(params: any): string {
  const label = params.data.name;
  const key = `${params.percent}%`;
  return template(label, { key, value: "" });
}

export function areaTooltip(params: any, title: string, nbrPoints: string, varType: string, data: number[]): string {
  // Format the tooltip content as needed
  const label = title;
  const value = { key: `${getShape('rectangle', 'green')} Recommended Area`, value: "" };
  let values;
  let categories = [];
  // Assuming params.value or params.data is the array of points for the mark area
  if(varType === 'cat'){
    values = data.length === 2 ? `from ${data.join(' to ')}` : data.slice(0,2).join(" - ");
    if(data.length > 2){
      for(let i=2; i < data.slice(0,10).length; i+=2){
        categories.push({ key: '', value: i+1 < data.length ? data[i] + " - " + data[i+1] : data[i]})
      }
      if(data.length > 10){
        categories.push({key: '', value: 'other...'})
      }
    }
  }else{
    const startVal = params.data.coord[0][0];
    const endVal = params.data.coord[1][0];
    values = `from ${formatNumbers(startVal)} to ${formatNumbers(endVal)}`;
  }
  const value2 = { key: `Variable values`, value:  values};
  const value3 = { key: `Points`, value: formatNumbers(nbrPoints) };
  if(categories.length){
    return template(label, value, value2, ...categories, value3);
  }
  return template(label, value, value2, value3);

}

function analysisTooltip(params: any, title: string, chartType: string, variableType: string): string {
  const label = title;
  const isExpert = chartType === "expert";
  if(params.seriesName !== "Number of points"){
    const valueColor = params.data[4] === 1 ? 'green' : 'red';
    const valueShape = params.data[4] === 1 ? getShape(!isExpert ? 'circle' : 'rectangle', valueColor) : getShape(!isExpert ? 'circle' : 'rectangle', valueColor);
    const valueText = params.data[4] === 1 ? 'OK' : 'NOK';
    const value = { key: `${valueShape} ${valueText}`, value: "" };
    const value2 = variableType === 'num' ? { key: `Variable values`, value: `${formatNumbers(params.data[5])} to ${formatNumbers(params.data[6])}` } : { key: `Variable value`, value: params.data[5] };
    const value3 = { key: `NOK rate `, value: `${formatNumbers(isExpert ? params.data[8]: params.data[1])}%` };
    const value4 = { key: `Points `, value: `${formatNumbers(params.data[3])}` };
    const value5 = { key: `Points % `, value: `${formatNumbers(params.data[7])}%` };
    return template(label, value, value2, value3, value4, value5);
  }else{
    const value = { key: `Points `, value: `${formatNumbers(params.data[1])}` };
    const value2 = { key: `Points % `, value: `${formatNumbers(params.data[2])}%` };
    return template(params.seriesName, value, value2);

  }
}

export function tooltipOptions(chartType: ChartType, title?: string, extra?: string, variableType?: string, data?: number[]) {
  return {
    trigger: 'item',
    formatter: function (params: any) {
      switch (chartType) {
        case ChartType.TargetScatter:
          return scatterTooltip(params);
        case ChartType.Pie:
          return pieTooltip(params);
        case ChartType.AnalysisArea:
          return areaTooltip(params, title || '', extra!, variableType!, data!);
        case ChartType.AnalysisScatter:
          return analysisTooltip(params, title || '', extra || 'standard', variableType!);
        default:
          return '';
      }
    },
    backgroundColor: '#FFF',
    borderColor: 'transparent',
    borderWidth: 0,
    padding: [8, 12],
    textStyle: {
      color: 'black',
    },
  };

};